/*==============================================================================
 * gb_common.c
 *------------------------------------------------------------------------------
 *
 * Copyright 2009-2012 Universitat Jaume I
 *
 * This file is part of GPUBenchmark.
 *
 * GPUBenchmark is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * GPUBenchmark is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GPUBenchmark.  If not, see <http://www.gnu.org/licenses/>.
 *==============================================================================
 */

#include <gb_common.h>

#include <getopt.h>
#include <string.h>
#include <stdarg.h>
#include <sys/time.h>

#include "../Config.h"

/**
   The program name, it will be set by getopts() from argv[0].
 */
char *program_name=0;


/**
   Gets the options from the command line and fills the given option
   struct. It also initializes the program_name string that will be
   used by the error() and warning() functions.

   @param [in]  argc       Count of the arguments supplied to the program
   @param [in]  **argv     Array of pointers to the strings which are those arguments
   @param [out] *option    Struct to be filled with the given option
 */
void getopts(int argc, char **argv, struct options_t *options)
{

  // Defaults
  (*options).device=0;           // Id of CUDA device to be used
  (*options).device2=-1;         // Id of second CUDA device to be used
  (*options).device_info=0;      // Show device information?
  (*options).m=-1;               // Rows and cols of matrices A, B and C in gemm and gemv
  (*options).n=-1;
  (*options).k=-1;
  (*options).rows=-1;            // Rows and cols in tests that use only one matrix
  (*options).cols=-1;
  (*options).padding=0;          // Do not use padding.
  (*options).rep=10;             // Number of repetitions of each test
  (*options).use_simple=1;       // Whether to use simple or double precision
  (*options).help=0;             // Show the usage help?
  (*options).output=stdout;      // Pointer to output FILE (stdout or a real file)
  (*options).verbosity=0;        // Verbosity level

  // Store the program name in the global variable program_name
  program_name=strdup(argv[0]);

  // Get the options from the command line
  int c;
  while (1) {
    int option_index = 0;

    static struct option long_options[] = {
      {"m",           required_argument, 0,  'm' },
      {"n",           required_argument, 0,  'n' },
      {"k",           required_argument, 0,  'k' },
      {"cols",        required_argument, 0,   0  },
      {"device",      required_argument, 0,   0  },
      {"device2",     required_argument, 0,   0  },
      {"device-info", no_argument,       0,   0  },
      {"double",      no_argument,       0,   0  },
      {"output",      required_argument, 0,  'o' },
      {"padding",     no_argument,       0,  'p' },
      {"rep",         required_argument, 0,  'r' },
      {"rows",        required_argument, 0,   0  },
      {"verbose",     no_argument,       0,  'v' },
      {"help",        no_argument,       0,  'h' },
      {0,             0,                 0,   0  }
    };

    c = getopt_long(argc, argv, "m:n:k:o:pr:vVh", long_options, &option_index);

    if (c == -1)
      break;

    switch (c)
      {
      case 0:
	if (strcmp("cols", long_options[option_index].name) == 0)
	  {
	    (*options).cols=atoi(optarg);
	    if ((*options).cols < 1)
	      error("The number of columns must be greater than 0.");
	  }
	else if (strcmp("rows", long_options[option_index].name) == 0)
	  {
	    (*options).rows=atoi(optarg);
	    if ((*options).rows < 1)
	      error("The number of rows must be greater than 0.");
	  }
	else if (strcmp("device", long_options[option_index].name) == 0)
	  {
	    (*options).device=atoi(optarg);
	    if ((*options).device < 0)
	      error("The CUDA device id must be greater than 0.");
	  }
        else if (strcmp("device2", long_options[option_index].name) == 0)
	  {
	    (*options).device2=atoi(optarg);
	    if ((*options).device2 < 0)
	      error("The CUDA device id must be greater than 0.");
	  }
	else if (strcmp("device-info", long_options[option_index].name) == 0)
	  {
	    (*options).device_info=1;
	  }
	else if (strcmp("double", long_options[option_index].name) == 0)
	  {
	    (*options).use_simple=0;
	  }
	else
	  {
	    printf("Warning: option %s", long_options[option_index].name);
	    if (optarg)
	      printf(" with arg %s", optarg);
	    printf(" is not supported.\n");
	  }
	break;

      case 'm':
	(*options).m=atoi(optarg);
	if ((*options).m < 1)
	  error("The value of m must be greater than 0.");
	break;

      case 'n':
	(*options).n=atoi(optarg);
	if ((*options).n < 1)
	  error("The value of n must be greater than 0.");
	break;

      case 'k':
	(*options).k=atoi(optarg);
	if ((*options).m < 1)
	  error("The value of k must be greater than 0.");
	break;

      case 'o':
	(*options).output=fopen(optarg, "w");
	break;

      case 'p':
	(*options).padding=1;
	break;

      case 'r':
	(*options).rep=atoi(optarg);
	if ((*options).rep<1)
	  error("The number of test repetitions must be greater than 0.");
	break;

      case 'v':
	(*options).verbosity++;
	break;

      case 'V':
	printf("GPUBenchmark version %d.%d (%s)\n", VERSION_MAJOR, VERSION_MINOR, program_name);
	exit(0);

      case '?':  /* fall-through is intentional */
      case 'h':
	(*options).help=1;
	break;

      default:
	warning("getopt returned unsupported character '%c'.", c);
      }
  }

  if (optind < argc) {
    printf("Warning: the next non-option ARGV-elements are not processed: ");
    while (optind < argc)
      printf("%s ", argv[optind++]);
    printf("\n");
  }

}

/**
   Prints and error message and exits with -1.
*/
void error(char *msg, ...)
{
  va_list argp;
  if (program_name)
    fprintf(stderr, "%s: ", program_name);
  else
    fprintf(stderr, "Error: ");
  va_start(argp, msg);
  vfprintf(stderr, msg, argp);
  va_end(argp);
  fprintf(stderr, "\n");
  exit(-1);
}

/**
   Prints and error message but does not exits.
*/
void no_exit_error(char *msg, ...)
{
  va_list argp;
  if (program_name)
    fprintf(stderr, "%s: ", program_name);
  else
    fprintf(stderr, "Error: ");
  va_start(argp, msg);
  vfprintf(stderr, msg, argp);
  va_end(argp);
  fprintf(stderr, "\n");
}

/**
   Prints a warning message.
*/
void warning(char *msg, ...)
{
  va_list argp;
  if (program_name)
    fprintf(stderr, "%s: warning: ", program_name);
  else
    fprintf(stderr, "Warning: ");
  va_start(argp, msg);
  vfprintf(stderr, msg, argp);
  va_end(argp);
  fprintf(stderr, "\n");
}


/**
   Utility function that compares two floats.
*/
int compare_float(const void *a, const void *b)
{
  float temp = *(float*)a - *(float*)b;
  if (temp > 0)       return  1;
  else if (temp < 0)  return -1;
  else                return  0;
}


/**
   Utility function that compares two doubles.
*/
int compare_double(const void *a, const void *b)
{
  double temp = *(double*)a - *(double*)b;
  if (temp > 0)       return  1;
  else if (temp < 0)  return -1;
  else                return  0;
}


/**
   Utility function that returns the mean of a vector of doubles.

   @param [in,out]  vector The vector from which the mean has to be computed.
                           The original vector will be sorted by this function.
   @param [in]      size   The size of the vector.
   @returns                The mean value.
*/
double mean(double *vector, int size)
{
  qsort(vector, size, sizeof(double), compare_double);
  return vector[size/2];
}


/**
   Utility function that returns the difference in seconds between two time values.

   @param [in]  t1  timeval struct with the last time value.
   @param [in]  t0  timeval struct with the first time value.
   @returns         The difference in seconds between t1 and t0.
*/
double timeval_diff(struct timeval t1, struct timeval t0)
{
  return (t1.tv_sec - t0.tv_sec) + (t1.tv_usec - t0.tv_usec)*1.0e-6;
}
