/*==============================================================================
 * gb_gputogpu.cu - measures the transfer speed between two GPUs memories
 *------------------------------------------------------------------------------
 *
 * Copyright 2009-2012 Universitat Jaume I
 *
 * This file is part of GPUBenchmark.
 *
 * GPUBenchmark is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * GPUBenchmark is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GPUBenchmark.  If not, see <http://www.gnu.org/licenses/>.
 *==============================================================================
 */


/* 
 * This test measures the transfer speed between two GPUs memory with
 * GPUDirect and compare the result with using the CPU's memory in the
 * transfer.  The measurement is done obtaining the time spent in
 * copying a matrix from one memory to other.
 *
 * The size of the matrix to copied, and the number of repetitions of
 * the copy are specified by the user at runtime.
 */


#include "gb_gputogpu.h"


/**
   Runs the gputogpu microbenchmark with the parameters specified
   by the supplied options.

   @param [in]  argc       Count of the arguments supplied to the program.
   @param [in]  **argv     Array of pointers to the strings which are those arguments.
   @return 0 if the test passed correctly, other number otherwise.
*/
int main(int argc, char** argv)
{
  struct options_t options; // Command line options struct
  int i;
  int num_devices;
  double *time_gpudirect_transfer;
  double *time_memory_transfer;
  double mean_time_gpudirect_transfer;
  double mean_time_memory_transfer;

  // Get options from command line
  getopts(argc, argv, &options);

  if (options.help)
    gputogpuHelp();

  if (options.device_info)
    showProperties(options.output, options.device);

  // Check the number of devices
  cudaGetDeviceCount(&num_devices);
  if (num_devices<2)
    error("This test can only be run in machines with more than one GPU device.");
      
  // Check the required parameters
  if (options.rows == -1 || options.cols == -1)
    error("The rows and cols parameters must be specified.");

  if (options.verbosity)
    fprintf(options.output,
            "\n"
            "Size of matrix: %.2f KiB\n",
	    (options.cols * options.rows * sizeof (float)) / 1024.0);

  time_gpudirect_transfer = (double *) malloc(options.rep * sizeof (double));
  time_memory_transfer = (double *) malloc(options.rep * sizeof (double));

  if (options.device2 == -1) options.device2 = (options.device == 0) ? 1 : 0;

  for (i=0; i<options.rep; i++) 
    {
      if (gputogpu(options.device, options.device2, options.rows, options.cols, &time_gpudirect_transfer[i], &time_memory_transfer[i]) == -1)
	error("The gputogpu test %d/%d failed.", i + 1, options.rep);
    }

  // Compute mean cpu and gpu times
  mean_time_gpudirect_transfer = mean(time_gpudirect_transfer, options.rep);
  mean_time_memory_transfer = mean(time_memory_transfer, options.rep);


  if (options.verbosity)
    {
      fprintf(options.output,
	      "\n"
	      "GPUDirect transfer time:  %7.4f ms\n"
	      "Memory transfer time:     %7.4f ms\n"
	      "GPUDirect transfer speed: %10.4f MiB/s\n"
	      "Memory transfer speed:    %10.4f MiB/s\n",
	      mean_time_gpudirect_transfer*1000.0,
	      mean_time_memory_transfer*1000.0,
	      options.rows*options.cols*sizeof(float)/MIB/mean_time_gpudirect_transfer,
	      options.rows*options.cols*sizeof(float)/MIB/mean_time_memory_transfer);
    }
  else
    {
      fprintf(options.output,
	      //12345678.12345678901234.12345678901234567890.12345678901234.12345678901234567890.
	      "#   M       N       Size      GPU1->GPU2 G1->MM->G2 GPU1->GPU2 G1->MM->G2 \n"
	      "#                   (MiB)        (ms)       (ms)      (MiB/s)    (MiB/s)  \n"
	      "#-------+-------+------------+----------+----------+----------+----------+\n"
	      " %7d %7d %12.2f %10.2f %10.2f %10.2f %10.2f\n"
	      "",
	      options.rows, options.cols,
	      options.rows*options.cols*sizeof(float)/MIB,
	      mean_time_gpudirect_transfer*1000.0,
	      mean_time_memory_transfer*1000.0,
	      options.rows*options.cols*sizeof(float)/MIB/mean_time_gpudirect_transfer,
	      options.rows*options.cols*sizeof(float)/MIB/mean_time_memory_transfer);
    }

  // Exit main
  fprintf(options.output, "\n");
  fclose(options.output);
  free(time_gpudirect_transfer);
  free(time_memory_transfer);
  cudaDeviceReset();
  return 0;
}

/**
   Measures the time of copying a matrix from a GPU memory to another
   GPU memory using gpuDirect.

   @param [in]  dev1                  The source GPU.
   @param [in]  dev2                  The destination GPU.
   @param [in]  rows                  The number of rows of the matrix to be copied.
   @param [in]  cols                  The number of cols of the matrix to be copied.
   @param [out] t_gpudirect_transfer  Time spent in doing the copy with GPUDirect (in seconds).
   @param [out] t_memory_transfer     Time spent in doing the copy using main memory (in seconds).
   @return 0 if everything went ok, other number otherwise.
*/
double gputogpu(int dev1, int dev2, int rows, int cols, double *t_gpudirect_transfer, double *t_memory_transfer)
{
  int i, j;
  float* matrix_mm;
  float* matrix_gpu1;
  float* matrix_gpu2;
  cudaEvent_t start, end;
  float float_time;
  bool error=false;

  // Allocate space in main memory for the matrix
  matrix_mm = (float *) malloc(rows * cols * sizeof (float));
  if (!matrix_mm)
    {
      no_exit_error("Allocation of matrix in main memory failed.");
      return -1;
    }

  // Fill the matrix with arbitrary values
  for (i = 0; i < rows; i++)
    for (j = 0; j < cols; j++)
      matrix_mm[i * cols + j] = i*cols+j;


  // Allocate space in GPUs memory
  cudaSetDevice(dev1);
  if (cudaMalloc((void **)&matrix_gpu1, rows*cols*sizeof(float)) != cudaSuccess)
    {
      no_exit_error("Allocation of memory in GPU %d failed.", dev1);
      free(matrix_mm);
      return -1;
    }
  cudaSetDevice(dev2);
  if (cudaMalloc((void **)&matrix_gpu2, rows*cols*sizeof(float)) != cudaSuccess)
    {
      no_exit_error("Allocation of memory in GPU %d failed.", dev2);
      free(matrix_mm);
      cudaSetDevice(dev1);
      cudaFree(matrix_gpu1);
      return -1;
    }

  // Copy the matrix from main memory to GPU1
  cudaSetDevice(dev1);
  if (cudaMemcpy(matrix_gpu1, matrix_mm, rows*cols*sizeof(float), cudaMemcpyHostToDevice) != cudaSuccess)
    {
      no_exit_error("Copying the matrix in GPU %d from main memory failed.", dev1);
      free(matrix_mm);
      cudaFree(matrix_gpu1);
      cudaSetDevice(dev2);
      cudaFree(matrix_gpu2);
      return -1;
    }

  // Create start and end events in dev1
  cudaSetDevice(dev1);
  cudaEventCreate(&start);
  cudaEventCreate(&end);


  // ======================================================================
  // Transfer between the two GPUs using GPUDirect
  // ----------------------------------------------------------------------
  cudaDeviceSynchronize();
  cudaEventRecord(start, 0);
  cudaMemcpyPeer(matrix_gpu2, dev2, matrix_gpu1, dev1, rows*cols*sizeof(float));
  cudaEventRecord(end, 0);
  cudaEventSynchronize(end);

  // Get the elapsed time (in ms) between "start" and "end" events
  cudaEventElapsedTime(&float_time, start, end);

  // Convert time to seconds and store it in t_gpudirect_transfer
  *t_gpudirect_transfer=(double)float_time/1000.0;
  // ======================================================================


  // ======================================================================
  // Transfer between the two GPUs using main memory
  // ----------------------------------------------------------------------
  cudaDeviceSynchronize();
  cudaEventRecord(start, 0);
  // Copy the matrix from GPU1 to main memory
  cudaSetDevice(dev1);
  if (cudaMemcpy(matrix_mm, matrix_gpu1, rows*cols*sizeof(float), cudaMemcpyDeviceToHost) != cudaSuccess)
    {
      no_exit_error("Copying the matrix in main memory from device %d failed.", dev1);
      free(matrix_mm);
      // dev1 cleanup
      cudaSetDevice(dev1);
      cudaFree(matrix_gpu1);
      cudaEventDestroy(start);
      cudaEventDestroy(end);
      // dev2 cleanup
      cudaSetDevice(dev2);
      cudaFree(matrix_gpu2);
      return -1;
    }
  // Copy the matrix from main memory to GPU2
  cudaSetDevice(dev2);
  if (cudaMemcpy(matrix_gpu2, matrix_mm, rows*cols*sizeof(float), cudaMemcpyHostToDevice) != cudaSuccess)
    {
      no_exit_error("Copying the matrix in device %d from main memory failed.", dev2);
      free(matrix_mm);
      // dev1 cleanup
      cudaSetDevice(dev1);
      cudaFree(matrix_gpu1);
      cudaEventDestroy(start);
      cudaEventDestroy(end);
      // dev2 cleanup
      cudaSetDevice(dev2);
      cudaFree(matrix_gpu2);
      return -1;
    }
  cudaSetDevice(dev1);
  cudaEventRecord(end, 0);
  cudaEventSynchronize(end);

  // Get the elapsed time (in ms) between "start" and "end" events
  cudaEventElapsedTime(&float_time, start, end);

  // Convert time to seconds and store it in t_memory_transfer
  *t_memory_transfer=(double)float_time/1000.0;
  // ======================================================================


  // Free the memories
  free(matrix_mm);
  cudaSetDevice(dev1);
  if (cudaFree(matrix_gpu1) != cudaSuccess)
    {
      no_exit_error("Freeing GPU %d memory failed.", dev1);
      error=true;
    } 
  cudaSetDevice(dev2);
  if (cudaFree(matrix_gpu2) != cudaSuccess)
    {
      no_exit_error("Freeing GPU %d memory failed.", dev2);
      error=true;
    }

  // Destroy "start" and "end" events
  cudaEventDestroy(start);
  cudaEventDestroy(end);

  // Return
  if (error)
    return -1;
  else
    return 0;
}


/**
   Show usage information
*/
void gputogpuHelp() {
  printf("Usage: ./gb_gputogpu [OPTION]..\n"
	 "Measures the transfer speed between two GPUs with GPUDirect, and\n"
	 "without it.\n"
	 "\n"
	 "Required options:\n"
	 "      --rows=NUM       number of rows of the matrix to be copied\n"
	 "      --cols=NUM       number of columns of the matrix to be copied\n"
	 "\n"
	 "Other options:\n"
	 "      --device=NUM     use this CUDA device as the source (default: 0)\n"
	 "      --device2=NUM    use this CUDA device as the destination (default: 1)\n"
	 "      --device-info    show information about the source device and exit\n"
	 "      --rep=NUM        number of repetitions of the test (default: 10)\n"
	 "  -v, --verbose        be more verbose\n"
	 "  -o, --output=FNAME   send the output to the specified file\n"
	 "      --help           display this help and exit\n"
	 "\n");
  exit(0);
}

