/*==============================================================================
 * gb_gputoram.cu - measures the transfer speed between main and GPU memories
 *------------------------------------------------------------------------------
 *
 * Copyright 2009-2012 Universitat Jaume I
 *
 * This file is part of GPUBenchmark.
 *
 * GPUBenchmark is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * GPUBenchmark is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GPUBenchmark.  If not, see <http://www.gnu.org/licenses/>.
 *==============================================================================
 */


/* 
 * This microbenchmark measures the transfer speed between the
 * computer's main memory and the GPU's memory, and viceversa.
 *
 * The measurement is done obtaining the time spent in copying a
 * matrix from one memory to other.
 *
 * The size of the matrix to copied, and the number of repetitions of
 * the copy are specified by the user at runtime.
 */

#include "gb_gputoram.h"


/**
   Runs the gputoram microbenchmark with the parameters specified
   by the supplied options.

   @param [in]  argc       Count of the arguments supplied to the program.
   @param [in]  **argv     Array of pointers to the strings which are those arguments.
   @return 0 if the test passed correctly, other number otherwise.
*/
int main(int argc, char** argv)
{
  struct options_t options; // Command line options struct
  int i;
  int sizeofelem;
  double *time_mem2gpu;
  double *time_gpu2mem;
  double mean_mem2gpu_speed;
  double mean_gpu2mem_speed;
  double size_in_kib;

  // Get options from command line
  getopts(argc, argv, &options);

  if (options.help)
    gputoramHelp();

  if (options.device_info)
    showProperties(options.output, options.device);

  // Check required parameters
  if (options.rows==-1 || options.cols==-1)
    error("The number of rows and columns must be specified.");

  // Set sizeofelem
  if (options.use_simple)
    sizeofelem=sizeof(float);
  else
    sizeofelem=sizeof(double);

  // Compute size of matrix in KiB
  size_in_kib=((double)options.rows*options.cols*sizeofelem)/1024.0;

  // Display test information
  if (options.verbosity)
    fprintf(options.output, "\nSize of matrix: %.2f KiB\n", size_in_kib);

  // Time vectors
  time_mem2gpu=(double *) malloc(options.rep*sizeof(double));
  time_gpu2mem=(double *) malloc(options.rep*sizeof(double));

  // Run the tests options.rep times
  for (i=0; i<options.rep; i++)
    {
      if (gputoram(options.device, options.rows, options.cols, sizeofelem, options.padding, &time_mem2gpu[i], &time_gpu2mem[i]) == -1)
	error("The test %d/%d failed.", i+1, options.rep);
    }

  // Compute mean read and write speeds
  mean_mem2gpu_speed=size_in_kib/1024.0/mean(time_mem2gpu, options.rep);
  mean_gpu2mem_speed=size_in_kib/1024.0/mean(time_gpu2mem, options.rep);
  
  // Display the mean transfer speeds
  if (options.verbosity)
    {
      fprintf(options.output,
	      "\n"
	      "Transfer speed (MM->GPU): %10.4f MiB/s\n"
	      "Transfer speed (GPU->MM): %10.4f MiB/s\n",
	      mean_mem2gpu_speed, mean_gpu2mem_speed);
    }
  else
    {
      fprintf(options.output,
	      //1234567890123.123456789012345.123456789012345.
	      "#   M       N     Size (MiB)  GPU->MM (MiB/s) MM->GPU (MiB/s) \n"
	      "#-------+-------+------------+---------------+---------------+\n"
	      " %7d %7d %12.2f %15.2f %15.2f\n"
	      "",
	      options.rows, options.cols,
	      options.rows*options.cols*sizeofelem/MIB,
	      mean_gpu2mem_speed,
	      mean_mem2gpu_speed);
    }

  // Exit main
  fprintf(options.output, "\n");
  fclose(options.output);
  free(time_mem2gpu);
  free(time_gpu2mem);
  cudaDeviceReset();
  return 0;
}


/**
   Measures the time required to copy a matrix of rows x cols elements
   from GPU to main memory, and viceversa.

   @param [in]  device        GPU device that is going to be used.
   @param [in]  rows          The number of rows of the matrix to be copied.
   @param [in]  cols          The number of cols of the matrix to be copied.
   @param [in]  sizeofelem    The size in bytes of each matrix element.
   @param [in]  padding       Whether to use padding or not to allocate the matrix.
   @param [out] time_mem2gpu  Time spent copying the matrix from main memory to GPU.
   @param [out] time_gpu2mem  Time spent copying the matrix from GPU to main memory.
   @return 0 if everything went ok, other number otherwise.
*/
int gputoram(int device, int rows, int cols, int sizeofelem, bool padding, double *time_mem2gpu, double *time_gpu2mem)
{
  int i, j;
  void *matrix_mm;
  void *matrix_gpu;
  size_t pitch=0;

  cudaEvent_t start, end;
  float float_time;

  // Allocate space in main memory for the matrix
  if (padding)
    {
      if (cudaMallocHost((void **)&matrix_mm, rows*cols*sizeofelem) != cudaSuccess)
	{
	  no_exit_error("Allocation of matrix in main memory with padding failed.");
	  return -1;
	}
    }
  else
    {
      matrix_mm= (void *) malloc(rows*cols*sizeofelem);
      if (!matrix_mm)
	{
	  no_exit_error("Allocation of matrix in main memory failed.");
	  return -1;
	}
    }
  
  // Fill the matrix with arbitrary values
  if (sizeofelem==sizeof(float))
    {
      for(i=0; i<rows; i++)
	for(j=0; j<cols; j++)
	  ((float *)matrix_mm)[i*cols+j]= i*cols+j;
    }
  else 
    {
      for(i=0; i<rows; i++)
	for(j=0; j<cols; j++)
	  ((double *)matrix_mm)[i*cols+j]= i*cols+j;
    }

  // Set CUDA device
  cudaSetDevice(device);

  // Allocate space in GPU memory
  if (padding)
    {
      if (cudaMallocPitch((void **)&matrix_gpu, &pitch, cols*sizeofelem, rows) != cudaSuccess)
	{
	  no_exit_error("Allocation of matrix in GPU memory failed.");
	  if (cudaFreeHost(matrix_mm)!= cudaSuccess)
	    no_exit_error("Freeing host memory failed.");
	  return -1;
	}
    }
  else
    {
      if (cudaMalloc((void **)&matrix_gpu, rows*cols*sizeofelem) != cudaSuccess)
	{
	  no_exit_error("Allocation of matrix in GPU memory failed."); 
	  free(matrix_mm);
	  return -1;
	}
    }
  
  // Create start and end events
  cudaEventCreate(&start);
  cudaEventCreate(&end);


  // ======================================================================
  // Main memory to GPU
  // ----------------------------------------------------------------------

  // Record "start" event
  cudaDeviceSynchronize();
  cudaEventRecord(start, 0);
  
  // Copy the matrix from main memory to GPU
  if (padding)
    {
      if (cudaMemcpy2D(matrix_gpu, pitch, matrix_mm,
		       cols*sizeofelem, cols*sizeofelem,
		       rows, cudaMemcpyHostToDevice) != cudaSuccess)
	{
	  no_exit_error("Copying the matrix to GPU memory with cudaMemcpy2D failed.");
	  cudaFreeHost(matrix_mm);
	  cudaFree(matrix_gpu);
	  cudaEventDestroy(start);
	  cudaEventDestroy(end);
	  return -1;
	}
    }
  else
    {
      if (cudaMemcpy(matrix_gpu, matrix_mm, rows*cols*sizeofelem,
		     cudaMemcpyHostToDevice) != cudaSuccess)
	{
	  no_exit_error("Copying the matrix to GPU memory with cudaMemcpy failed\n");
	  free(matrix_mm);
	  cudaFree(matrix_gpu);
	  cudaEventDestroy(start);
	  cudaEventDestroy(end);
	  return -1;
	}
    }

  // Record "end" event and synchronize
  cudaEventRecord(end, 0);
  cudaEventSynchronize(end);

  // Get the elapsed time (in ms) between "start" and "end" events
  cudaEventElapsedTime(&float_time, start, end);

  // Convert time to seconds and store it in time_mem2gpu
  *time_mem2gpu=(double)float_time/1000.0;

  // ======================================================================


  // Free actual reserved memory and reallocate space for another copy
  if (padding) 
    {
      if (cudaFreeHost(matrix_mm)!= cudaSuccess)
	{
	  no_exit_error("Freeing host memory for another copy failed.");
	  if (cudaFree(matrix_gpu) != cudaSuccess)
	    no_exit_error("Freeing GPU memory failed.");
	  cudaEventDestroy(start);
	  cudaEventDestroy(end);
	  return -1;
	}
    }
  else
    {
      free(matrix_mm);
    }

  // Allocate space in main memory for the "GPU to main memory" copy
  if (padding)
    {
      if (cudaMallocHost((void **)&matrix_mm, rows*cols*sizeofelem) != cudaSuccess)
	{
	  no_exit_error("Allocation of matrix in main memory failed.");
	  cudaFree(matrix_gpu);
	  cudaEventDestroy(start);
	  cudaEventDestroy(end);
	  return -1;
	}
    }
  else
    {
      matrix_mm= (void *) malloc(rows*cols*sizeofelem);
      if (!matrix_mm)
	{
	  no_exit_error("Allocation of matrix in main memory failed.");
	  cudaFree(matrix_gpu);
	  cudaEventDestroy(start);
	  cudaEventDestroy(end);
	  return -1;
	}
    }


  // ======================================================================
  // GPU to main memory
  // ----------------------------------------------------------------------

  // Record "start" event
  cudaDeviceSynchronize();
  cudaEventRecord(start, 0);
  
  // Copy the matrix from GPU to main memory
  if (padding)
    {
      if (cudaMemcpy2D(matrix_mm, cols*sizeofelem, matrix_gpu,
		       pitch, cols*sizeofelem, rows, cudaMemcpyDeviceToHost) != cudaSuccess)
	{
	  no_exit_error("Copying the matrix to main memory with cudaMemcpy2D failed.");
	  cudaFreeHost(matrix_mm);
	  cudaFree(matrix_gpu);
	  cudaEventDestroy(start);
	  cudaEventDestroy(end);
	  return -1;
	}
    }
  else
    {
      if (cudaMemcpy(matrix_mm, matrix_gpu, rows*cols*sizeofelem,
		     cudaMemcpyDeviceToHost) != cudaSuccess)
	{
	  no_exit_error("Copying the matrix to main memory with cudaMemcpy failed.");
	  free(matrix_mm);
	  cudaFree(matrix_gpu);
	  cudaEventDestroy(start);
	  cudaEventDestroy(end);
	  return -1;
	}
    }

  // Record "end" event and synchronize
  cudaEventRecord(end, 0);
  cudaEventSynchronize(end);

  // Get the elapsed time (in ms) between "start" and "end" events
  cudaEventElapsedTime(&float_time, start, end);

  // Convert time to seconds and store it in time_gpu2mem
  *time_gpu2mem=(double)float_time/1000.0;

  // ======================================================================


  // Free the memories
  if (cudaFree(matrix_gpu) != cudaSuccess)
    {
      no_exit_error("Freeing GPU memory failed.");
      return -1;
    }

  if(padding)
    {
      if(cudaFreeHost(matrix_mm)!= cudaSuccess)
	{
	  no_exit_error("Freeing host memory failed.");
	  return -1;
	}
    }
  else
    free(matrix_mm);
  
  // Destroy "start" and "end" events
  cudaEventDestroy(start);
  cudaEventDestroy(end);

  // Return
  return 0;
}



/**
   Show usage information
*/
void gputoramHelp(void)
{
  printf("Usage: ./gb_gputoram [OPTION]..\n"
	 "Measures the transfer speed between GPU and main memory, and vice versa.\n"
	 "\n"
	 "Required options:\n"
	 "      --rows=NUM       number of rows of the matrix to be copied\n"
	 "      --cols=NUM       number of columns of the matrix to be copied\n"
	 "\n"
	 "Other options:\n"
	 "      --device=NUM     use the CUDA device with id=NUM (default: 0)\n"
	 "      --device-info    show information about the given device and exit\n"
	 "      --double         run the test using double precision elements\n"
	 "  -p, --padding        allocate the required memory using padding\n"
	 "      --rep=NUM        number of repetitions of the test (default: 10)\n"
	 "  -v, --verbose        be more verbose\n"
	 "  -o, --output=FNAME   send the output to the specified file\n"
	 "      --help           display this help and exit\n"
	 "\n");
  exit(0);
}
