/*==============================================================================
 * gb_hdtoram.c - measures the transfer speed between hard disk and main memory
 *------------------------------------------------------------------------------
 *
 * Copyright 2009-2012 Universitat Jaume I
 *
 * This file is part of GPUBenchmark.
 *
 * GPUBenchmark is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * GPUBenchmark is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GPUBenchmark.  If not, see <http://www.gnu.org/licenses/>.
 *==============================================================================
 */


/* 
 * This test measures the transfer speed between hard disk and main
 * memory.
 *
 * The measurement is done catching the time spent reading and writing
 * a file that contains a matrix or a vector using fread and fwrite
 * functions
 *
 * The size of the matrix/vector, and the number of repetitions of the
 * operation are specified by the user at runtime.
 *
 * This test is made both for single-precision and double-precision
 * numbers.
 *
 */

#include "gb_hdtoram.h"

#define tmp_fname "tmp_gb_hdtoram.dat"


/**
   Runs the hdtoram microbenchmark with the parameters specified by the
   supplied command line options.

   @param [in]  argc       Count of the arguments supplied to the program.
   @param [in]  **argv     Array of pointers to the strings which are those arguments.
   @return 0 if the test passed correctly, other number otherwise.
*/
int main(int argc, char **argv)
{
  struct options_t options; // Command line options struct
  int i;
  int sizeofelem;
  double *time_read;
  double *time_write;
  double mean_read_speed;
  double mean_write_speed;
  double size_in_kib;

  // Get options from command line
  getopts(argc, argv, &options);

  if (options.help)
    hdtoramHelp();

  // Check required parameters
  if (options.rows==-1 || options.cols==-1)
    error("The number of rows and columns must be specified.");

  // Set sizeofelem
  if (options.use_simple)
    sizeofelem=sizeof(float);
  else
    sizeofelem=sizeof(double);

  // Compute size of matrix in KiB
  size_in_kib=((double)options.rows*options.cols*sizeofelem)/1024.0;

  // Display test information
  if (options.verbosity)
    fprintf(options.output, "\nSize of matrix: %.2f KiB\n", size_in_kib);

  // Time vectors
  time_read=(double *) malloc(options.rep*sizeof(double));
  time_write=(double *) malloc(options.rep*sizeof(double));

  // Run the test options.rep times
  for (i=0; i<options.rep; i++)
    if (hdtoram(options.rows, options.cols, sizeofelem, &time_read[i], &time_write[i]) == -1)
      error("The test %d/%d failed.", i+1, options.rep);

  // Compute mean read and write speeds
  mean_read_speed=size_in_kib/1024.0/mean(time_read, options.rep);
  mean_write_speed=size_in_kib/1024.0/mean(time_write, options.rep);

  // Display the mean transfer speeds
  if (options.verbosity)
    {
      fprintf(options.output,
	      "\n"
	      "Read speed (HD->MM):  %10.4f MiB/s\n"
	      "Write speed (MM->HD): %10.4f Mib/s\n", 
	      mean_read_speed,
	      mean_write_speed);
    }
  else
    {
      fprintf(options.output,
	      //1234567890123.1234567890123456789.1234567890123456789.
	      "#   M       N     Size (MiB)   HD->MM (MiB/s)  MM->HD (MiB/s) \n"
	      "#-------+-------+------------+---------------+---------------+\n"
	      " %7d %7d %12.2f %15.2f %15.2f\n"
	      "",
	      options.rows, options.cols,
	      options.rows*options.cols*sizeofelem/MIB,
	      mean_read_speed,
	      mean_write_speed);
    }
  
  // Free time_read and time_write
  free(time_read);
  free(time_write);

  // Exit main
  fprintf(options.output, "\n");
  fclose(options.output);
  return 0;
}


/**
   Measures the time required for writing a matrix from main memory to
   a file on hard disk, and also the time required for reading the
   matrix from a file.

   @param [in]  rows          The number of rows of the matrix to be copied.
   @param [in]  cols          The number of cols of the matrix to be copied.
   @param [in]  sizeofelem    The size in bytes of each matrix element.
   @param [out] t_read Time spent reading a matrix from a file.
   @param [out] t_write Time spent writing a matrix to a file.
   @return 0 if everything went ok, other number otherwise.
*/
int hdtoram(int rows, int cols, int sizeofelem, double *t_read, double *t_write)
{
  int i, j, total_copied;

  void *matrix_mm;
  FILE *matrix_file;
  int fd;
  
  double tv_sec, tv_usec;
  struct timeval t0, t1;
  struct timezone z;
  
  //----------------------------------------------------------------------
  // Copy from memory to hard disk
  //----------------------------------------------------------------------

  // Allocate space in main memory for the matrix
  matrix_mm= (void *)malloc(rows*cols*sizeofelem);
  if (!matrix_mm)
    {
      no_exit_error("Allocation of matrix in main memory failed.");
      return -1;
    }
  
  // Fill matrix with arbitrary values (column-major indexing)
  if (sizeofelem==sizeof(float))
    for (j=0; j<cols; j++)
      for (i=0; i<rows; i++)
	((float *)matrix_mm)[IDX2C(i,j,rows)]= j*rows+i;
  else
    for (j=0; j<cols; j++)
      for (i=0; i<rows; i++)
	((double *)matrix_mm)[IDX2C(i,j,rows)]= j*rows+i;
    
  
  // Create a file to store the matrix
  matrix_file=fopen(tmp_fname, "wb");
  if(matrix_file == NULL)
    {
      no_exit_error("Creation of a file to store the matrix failed.");
      return -1;
    }

  // Get fd from the FILE pointer (it will be used to sync it)
  fd = fileno(matrix_file);

  // Capture current time of day (for MM->HD)
  gettimeofday(&t0, &z);

  // Copy the matrix fom main memory to file (HD)
  total_copied= fwrite(matrix_mm, sizeofelem, rows*cols, matrix_file);

  /*
   *   From "man fclose":
   *     Note that fclose() only flushes the user space buffers
   *     provided by the C library.  To ensure that the data is
   *     physically stored on disk the kernel buffers must be flushed
   *     too, for example, with sync(2) or fsync(2).
   *
   *   posix_fadvise usage from:
   *     "Improving Linux performance by preserving Buffer Cache State"
   *      http://insights.oetiker.ch/linux/fadvise/
   */

  // Sync the file, tell the OS that we no longer need it, and close it
  fsync(fd);
  posix_fadvise(fd, 0, 0, POSIX_FADV_DONTNEED); // Remove this file from the file system cache
  fclose(matrix_file);

  // Capture time of day after doing the storing operation
  gettimeofday(&t1, &z);

  // Compute the time difference between t1 and t0
  *t_write = timeval_diff(t1, t0);

  // Free memory
  free(matrix_mm);

  // Verify that the storage has been done correctly
  if (total_copied != rows*cols)
    {
      no_exit_error("Storing matrix in a file failed.");
      remove(tmp_fname);
      return -1;
    }


  //----------------------------------------------------------------------
  // Copy from hard disk to memory
  //----------------------------------------------------------------------

  // Reopen the file to read the matrix
  matrix_file=fopen(tmp_fname, "rb");
  if(matrix_file == NULL)
    {
      no_exit_error("Open the file with the matrix failed.");
      return -1;
    }

  // Allocate new space in main memory for the matrix
  matrix_mm= (void *)malloc(rows*cols*sizeofelem);

  // Capture current time of day (for HD->MM)
  gettimeofday(&t0, &z);
  
  // Read the matrix from file to main memory
  total_copied= fread(matrix_mm, sizeofelem, rows*cols, matrix_file);
  fclose(matrix_file);

  // Capture time of day after doing the reading operation
  gettimeofday(&t1, &z);

  // Compute time difference between t1 and t0
  *t_read = timeval_diff(t1, t0);

  // Free main memory remove the file
  free(matrix_mm);
  remove(tmp_fname);

  // Verify that the reading has been done correctly
  if(total_copied != rows*cols)
    {
      no_exit_error("Reading the matrix from a file failed.");
      return -1;
    }

  return 0;
}


/**
   Displays usage information
*/
void hdtoramHelp(void)
{
  printf("Usage: ./gb_hdtoram [OPTION]..\n"
	 "Measures the transfer speed between the hard disk and main memory.\n"
	 "\n"
	 "Required options:\n"
	 "      --rows=NUM       number of rows of the matrix to be copied\n"
	 "      --cols=NUM       number of columns of the matrix to be copied\n"
	 "\n"
	 "Other options:\n"
	 "      --double         runs the test using double precision elements\n"
	 "      --rep=NUM        number of repetitions of the test (default: 10)\n"
	 "  -v, --verbose        be more verbose\n"
	 "  -o, --output=FNAME   sends the output to the specified file\n"
	 "      --help           displays this help\n"
	 "\n");
  exit(0);
}
