/*==============================================================================
 * gb_matrixmatrix.cu - measures the computational power of the GPU and the CPU
 *------------------------------------------------------------------------------
 *
 * Copyright 2009-2012 Universitat Jaume I
 *
 * This file is part of GPUBenchmark.
 *
 * GPUBenchmark is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * GPUBenchmark is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GPUBenchmark.  If not, see <http://www.gnu.org/licenses/>.
 *==============================================================================
 */


/*
 * This microbenchmark measures the computational power of the CPU and
 * GPU using functions from CBLAS and CUBLAS libraries, respectively.
 *
 * The measurement is done obtaining the time spent by the operation:
 *
 *   C=alpha*A*B + beta*C,
 *
 * where A, B and C are matrices; and alpha and beta are two scalars.
 *
 * We can do this using gemm routines from CUBLAS (to do the
 * computation in the GPU) and CBLAS (to do the computation in the
 * CPU) libraries.
 *
 * The size of the matrices, and the number of repetitions of the
 * operation are specified by the user at runtime.
 *
 * This test is made both for single-precision and double precision
 * numbers. There are routines in CBLAS and CUBLAS supporting both
 * types.
 */


#include "gb_matrixmatrix.h"


/**
   Runs the matrixmatrix microbenchmark with the parameters specified
   by the supplied options.

   @param [in]  argc       Count of the arguments supplied to the program.
   @param [in]  **argv     Array of pointers to the strings which are those arguments.
   @return 0 if the test passed correctly, other number otherwise.
*/
int main(int argc, char** argv)
{
  struct options_t options; // Command line options struct
  int i;
  int sizeofelem;
  double *time_gpu;
  double *time_cpu;
  double mean_time_gpu;
  double mean_time_cpu;

  // Get options from command line
  getopts(argc, argv, &options);

  if (options.help)
    matrixmatrixHelp();

  if (options.device_info)
    showProperties(options.output, options.device);

  // Check the required parameters
  if (options.m==-1 || options.n==-1 || options.k==-1)
    error("The m, n, and k parameters must be specified.");

  // Set sizeofelem
  if (options.use_simple)
    sizeofelem=sizeof(float);
  else
    sizeofelem=sizeof(double);

  // Time vectors
  time_gpu=(double *) malloc(options.rep*sizeof(double));
  time_cpu=(double *) malloc(options.rep*sizeof(double));

  // Display matrix sizes
  if (options.verbosity)
    fprintf(options.output,
	    "\n"
	    "Size of matrix \"A\": %8.2f KiB\n"
	    "Size of matrix \"B\": %8.2f KiB\n"
	    "Size of matrix \"C\": %8.2f KiB\n\n",
	    (options.m*options.k*sizeofelem)/1024.0,
	    (options.k*options.n*sizeofelem)/1024.0,
	    (options.m*options.n*sizeofelem)/1024.0);

  // Run the tests options.rep times
  for (i=0; i<options.rep; i++)
    {
      if (matrixmatrix(options.device, options.m, options.n, options.k, sizeofelem, &time_cpu[i], &time_gpu[i]) == -1)
	error("The test %d/%d failed.", i+1, options.rep);
    }

  // Compute mean cpu and gpu times
  mean_time_cpu=mean(time_cpu, options.rep);
  mean_time_gpu=mean(time_gpu, options.rep);

  // @warning: '2.0' is used instead of '2' to overcome integer
  //           overflow when computing Gflops
  if (options.verbosity)
    {
      fprintf(options.output,
	      "CPU time (CBLAS):  %7.4f ms\n"
	      "GPU time (CUBLAS): %7.4f ms\n"
	      "GFLOPS CPU: %11.4f\n"
	      "GFLOPS GPU: %11.4f\n",
	      mean_time_cpu*1000.0,
	      mean_time_gpu*1000.0,
	      2.0*options.m*options.n*options.k*1.0e-9/mean_time_cpu,
	      2.0*options.m*options.n*options.k*1.0e-9/mean_time_gpu);
    } 
  else 
    {
      fprintf(options.output,
	      //1234567.1234567.1234567.1234567890123.1234567890123.1234567890123.1234567890123.
	      "#   M       N       K      CBLAS (ms)   CUBLAS (ms)   GFLOPS CPU    GFLOPS GPU\n"
	      "#-------+-------+-------+-------------+-------------+-------------+-------------+\n"
	      " %7d %7d %7d %13.2f %13.2f %13.2f %13.2f\n"
	      "",
	      options.m, options.n, options.k,
	      mean_time_cpu*1000.0,
	      mean_time_gpu*1000.0,
	      2.0*options.m*options.n*options.k*1.0e-9/mean_time_cpu,
	      2.0*options.m*options.n*options.k*1.0e-9/mean_time_gpu);
    }

  // Exit main
  fprintf(options.output, "\n");
  fclose(options.output);
  free(time_cpu);
  free(time_gpu);
  cudaDeviceReset();
  return 0;
}


/**
   Measures the time spent by the CPU and the GPU to do the operation
   \f$C=alpha*A*B+beta*C\f$, using the Sgemm() or Dgemm() functions of
   CBLAS and CUBLAS libraries, respectively.

   \em A, \em B and \em C are matrices, and \em alpha and \em beta are
   scalars.

   @param [in]  device       GPU device that is going to be used.
   @param [in]  m            The number of rows of matrices A and C.
   @param [in]  n            The number of columns of B and C.
   @param [in]  k	     The number of columns of A and rows of B.
   @param [in]  sizeofelem   The size in bytes of each matrix element.
   @param [out] time_cpu     Time spent by CPU to do the operation (in seconds).
   @param [out] time_gpu     Time spent by GPU to do the operation (in seconds).
   @return 0 if everything went ok, other number otherwise.
*/
int matrixmatrix(int device, int m, int n, int k, int sizeofelem, double *time_cpu, double *time_gpu)
{
  cudaError_t cudaStat;
  cublasStatus_t status;
  cublasHandle_t handle;
  bool error=false;
  int i, j;

  float s_alpha=2.0;
  float s_beta=3.0;
  double d_alpha=2.0;
  double d_beta=3.0;
  void *A_mm,  *B_mm,  *C_mm;
  void *A_gpu, *B_gpu, *C_gpu;
  void *C_result;

  double norm_C_result, norm_C_mm, result;

  cudaEvent_t start, end;
  float float_time;

  struct timeval t0, t1;
  struct timezone z;

  // Allocate space in main memory for CPU matrices
  A_mm= (void *) malloc(m*k*sizeofelem);
  B_mm= (void *) malloc(k*n*sizeofelem);
  C_mm= (void *) malloc(m*n*sizeofelem);
  C_result= (void *) malloc(m*n*sizeofelem);
  if (!A_mm || !B_mm || !C_mm || !C_result)
    {
      no_exit_error("Allocation of matrices in main memory failed.");
      free(A_mm);
      free(B_mm);
      free(C_mm);
      free(C_result);
      return -1;
    }

  // Fill matrices A, B, and C with arbitrary values (column-major indexing)
  if (sizeofelem==sizeof(float))
    {
      for (j=0; j<k; j++)
	for (i=0; i<m; i++)
	  ((float*)A_mm)[IDX2C(i,j,m)]= j*m+i;
      for (j=0; j<n; j++)
	for (i=0; i<k; i++)
	  ((float*)B_mm)[IDX2C(i,j,k)]= j*k+i;
      for (j=0; j<n; j++)
	for (i=0; i<m; i++)
	  ((float*)C_mm)[IDX2C(i,j,m)]= j*m+i;
    }
  else
    {
      for (j=0; j<k; j++)
	for (i=0; i<m; i++)
	  ((double*)A_mm)[IDX2C(i,j,m)]= j*m+i;
      for (j=0; j<n; j++)
	for (i=0; i<k; i++)
	  ((double*)B_mm)[IDX2C(i,j,k)]= j*k+i;
      for (j=0; j<n; j++)
	for (i=0; i<m; i++)
	  ((double*)C_mm)[IDX2C(i,j,m)]= j*m+i;
    }

  // Set CUDA device
  cudaSetDevice(device);

  // Allocate space in GPU memory for the matrices
  cudaStat = cudaMalloc((void**)&A_gpu, m*k*sizeofelem);
  if (cudaStat != cudaSuccess)
    {
      no_exit_error("Device memory allocation failed for matrix \"A\".");
      free(A_mm);
      free(B_mm);
      free(C_mm);
      free(C_result);
      return -1;
    }
  cudaStat = cudaMalloc((void**)&B_gpu, k*n*sizeofelem);
  if (cudaStat != cudaSuccess)
    {
      no_exit_error("Device memory allocation failed for matrix \"B\".");
      free(A_mm);
      free(B_mm);
      free(C_mm);
      free(C_result);
      cudaFree(A_gpu);
      return -1;
    }
  cudaStat = cudaMalloc((void**)&C_gpu, m*n*sizeofelem);
  if (cudaStat != cudaSuccess)
    {
      no_exit_error("Device memory allocation failed for matrix \"C\".");
      free(A_mm);
      free(B_mm);
      free(C_mm);
      free(C_result);
      cudaFree(A_gpu);
      cudaFree(B_gpu);
      return -1;
    }

  // Initialization of CUBLAS
  status = cublasCreate(&handle);
  if (status != CUBLAS_STATUS_SUCCESS)
    {
      no_exit_error("CUBLAS initialization failed.");
      return -1;
    }

  // Copy the matrices data fom main memory to GPU
  status= cublasSetMatrix(m, k, sizeofelem, A_mm, m, A_gpu, m);
  if (status != CUBLAS_STATUS_SUCCESS)
    {
      no_exit_error("Copying of matrix \"A\" to GPU memory failed.");
      error=true;
    }
  status= cublasSetMatrix(k, n, sizeofelem, B_mm, k, B_gpu, k);
  if (status != CUBLAS_STATUS_SUCCESS)
    {
      no_exit_error("Copying of matrix \"B\" to GPU memory failed.");
      error=true;
    }
  status= cublasSetMatrix(m, n, sizeofelem, C_mm, m, C_gpu, m);
  if (status != CUBLAS_STATUS_SUCCESS)
    {
      no_exit_error("Copying of matrix \"C\" to GPU memory failed.");
      error=true;
    }
  if (error) 
    {
      free(A_mm);
      free(B_mm);
      free(C_mm);
      free(C_result);
      cudaFree(A_gpu);
      cudaFree(B_gpu);
      cudaFree(C_gpu);
      return -1;
    }

  // Create start and end events
  cudaEventCreate(&start);
  cudaEventCreate(&end);


  // ======================================================================
  // CUBLAS operation
  // ----------------------------------------------------------------------

  // Record "start" event
  cudaDeviceSynchronize();
  cudaEventRecord(start, 0);

  // Do the gemm operation on GPU using cublas<T>gemm function
  if (sizeofelem==sizeof(float))
    status=cublasSgemm(handle, CUBLAS_OP_N, CUBLAS_OP_N, m, n, k, 
		       &s_alpha,  (float*)A_gpu, m,  (float*)B_gpu, k, &s_beta,  (float*)C_gpu, m);
  else
    status=cublasDgemm(handle, CUBLAS_OP_N, CUBLAS_OP_N, m, n, k, 
		       &d_alpha, (double*)A_gpu, m, (double*)B_gpu, k, &d_beta, (double*)C_gpu, m);

  if (status != CUBLAS_STATUS_SUCCESS)
    {
      if (sizeofelem==sizeof(float))
	no_exit_error("The cublasSgemm function failed.");
      else
	no_exit_error("The cublasDgemm function failed.");
      free(A_mm);
      free(B_mm);
      free(C_mm);
      free(C_result);
      cudaFree(A_gpu);
      cudaFree(B_gpu);
      cudaFree(C_gpu);
      cublasDestroy(handle);
      cudaEventDestroy(start);
      cudaEventDestroy(end);
      return -1;
    }

  // Record "end" event and synchronize
  cudaEventRecord(end, 0);
  cudaEventSynchronize(end);

  // Get the elapsed time (in ms) between "start" and "end" events
  cudaEventElapsedTime(&float_time, start, end);

  // Convert time to seconds and store it in time_gpu
  *time_gpu=(double)float_time/1000.0;

  // ======================================================================


  // Copy the result from GPU memory space to main memory space
  status = cublasGetMatrix(m, n, sizeofelem, C_gpu, m, C_result, m);
  if (status != CUBLAS_STATUS_SUCCESS)
    {
      no_exit_error("Copying of matrix \"C\" to CPU memory failed.");
      free(A_mm);
      free(B_mm);
      free(C_mm);
      free(C_result);
      cudaFree(A_gpu);
      cudaFree(B_gpu);
      cudaFree(C_gpu);
      cublasDestroy(handle);
      cudaEventDestroy(start);
      cudaEventDestroy(end);
      return -1;
    }


  // ======================================================================
  // CBLAS operation
  // ----------------------------------------------------------------------

  // Capture current time of day (for CBLAS <T>gemm)
  gettimeofday(&t0, &z);

  // Do the gemm operation on CPU using cblas_<T>gemm function
  if (sizeofelem==sizeof(float))
    cblas_sgemm(CblasColMajor, CblasNoTrans, CblasNoTrans, m, n, k,
		s_alpha,  (float*)A_mm, m,  (float*)B_mm, k, s_beta,  (float*)C_mm, m);
  else
    cblas_dgemm(CblasColMajor, CblasNoTrans, CblasNoTrans, m, n, k,
		d_alpha, (double*)A_mm, m, (double*)B_mm, k, d_beta, (double*)C_mm, m);

  // Capture time of day after doing the CBLAS <T>gemm operation
  gettimeofday(&t1, &z);

  // Compute the time difference between t1 and t0
  *time_cpu = timeval_diff(t1, t0);

  // ======================================================================


  // Compare the results obtained using CBLAS (CPU) and CUBLAS (GPU)
  if (sizeofelem==sizeof(float))
    {
      for (j=0; j<n; j++)
	for (i=0; i<m; i++)
	((float*)C_mm)[IDX2C(i,j,m)]=((float*)C_result)[IDX2C(i,j,m)]-((float*)C_mm)[IDX2C(i,j,m)];
      norm_C_result= cblas_snrm2(m*n, (float*)C_result, 1);
      norm_C_mm= cblas_snrm2(m*n, (float*)C_mm, 1);
    }
  else
    {
      for (j=0; j<n; j++)
	for (i=0; i<m; i++)
	((double*)C_mm)[IDX2C(i,j,m)]=((double*)C_result)[IDX2C(i,j,m)]-((double*)C_mm)[IDX2C(i,j,m)];
      norm_C_result= cblas_dnrm2(m*n, (double*)C_result, 1);
      norm_C_mm= cblas_dnrm2(m*n, (double*)C_mm, 1);
    }
  if (norm_C_result == 0.0)
    result= norm_C_mm;
  else
    result= norm_C_mm/norm_C_result;
  if ( (sizeofelem==sizeof(float) and result > FLT_EPSILON*10) or (sizeofelem==sizeof(double) and result > DBL_EPSILON*10 ) )
    warning("The difference between the results obtained with CBLAS and CUBLAS are too significative.");


  // Free the memories
  free(A_mm);
  free(B_mm);
  free(C_mm);
  free(C_result);
  cudaFree(A_gpu);
  cudaFree(B_gpu);
  cudaFree(C_gpu);

  // Destroy CUBLAS handle
  cublasDestroy(handle);

  // Destroy "start" and "end" events
  cudaEventDestroy(start);
  cudaEventDestroy(end);

  // Return
  return 0;
}


/**
   Show usage information
*/
void matrixmatrixHelp(void)
{
  printf("Usage: ./gb_matrixmatrix [OPTION]..\n"
	 "Measures the performance of gpu and cpu for the matrix matrix\n"
	 "multiplication operation using CUBLAS and CBLAS, respectivelly.\n"
	 "\n"
	 "Required options:\n"
	 "  -m, --m=NUM          number of rows of matrices A and C\n"
	 "  -n, --n=NUM          number of columns of B and C\n"
	 "  -k, --k=NUM          number of columns of A and rows of B\n"
	 "\n"
	 "Other options:\n"
	 "      --device=NUM     use  the CUDA device with id=NUM (default: 0)\n"
	 "      --device-info    show information about the given device and exit\n"
	 "      --double         run the test using double precision elements\n"
	 "      --rep=NUM        number of repetitions of the test (default: 10)\n"
	 "  -v, --verbose        be more verbose\n"
	 "      --dnfo           show information about the used device\n"
	 "  -o, --output=FNAME   send the output to the specified file\n"
	 "      --help           display this help and exit\n"
	 "\n");
  exit(0);
}
