#!/bin/bash

#===============================================================================
# gb_matrixmatrix_run.sh
#-------------------------------------------------------------------------------
#
# Copyright 2009-2012 Universitat Jaume I
#
# This file is part of GPUBenchmark.
#
# GPUBenchmark is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GPUBenchmark is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GPUBenchmark.  If not, see <http://www.gnu.org/licenses/>.
#===============================================================================


#===============================================================================
# Tests parameters
#-------------------------------------------------------------------------------
# matrixmatrix
#  - percent_nk are the % of m
matrixmatrix_m=(128 256 512 1024 2048 4096) # 8192 10240)
matrixmatrix_percent_nk=(25 50 75 100)
#===============================================================================


#===============================================================================
# Check that we are run from gb_run.sh
#-------------------------------------------------------------------------------
if [ -z "${GB_DIR}" ]; then
    echo "Error: this file must be called from 'gb_run.sh'"
fi
#===============================================================================


#===============================================================================
# matrixmatrix tests
#-------------------------------------------------------------------------------
echo "Running matrix per matrix tests..."

#-------------------------------------------------------------------------------
# Checks and initialization of variables
#-------------------------------------------------------------------------------
ELEM=$[ ${#matrixmatrix_m[@]} - ${ONE} ]
TEST=gb_matrixmatrix
TEST_EXE="${TEST}/${TEST}"
TEST_OUTPUT_DIR="${GB_OUTPUT_DIR}/${TEST}"
[ -d "${TEST_OUTPUT_DIR}" ] || mkdir -p "${TEST_OUTPUT_DIR}"

PLOTCPU="plot "
PLOTGPU="plot "
MM_PERCENTSNK=""
MM_SUBFIGURES=""
MM_REFSUBFIGURE=""
MM_TABLE=""

for i in $(seq 0 $[ ${#matrixmatrix_percent_nk[@]} - 1] )
do
    TEST_OUTPUT_FILE="${TEST_OUTPUT_DIR}/${TEST}_${i}.out"
    [ -e "${TEST_OUTPUT_FILE}" ] && rm -f "${TEST_OUTPUT_FILE}"

    #---------------------------------------------------------------------------
    # Actual running of test
    #---------------------------------------------------------------------------
    echo -n "Running matrix per matrix test ($[ ${i} + 1 ]/${#matrixmatrix_percent_nk[@]}): "
    progress_bar ${ELEM}
    for j in $(seq 0 ${ELEM})
    do
         n=$[ ${matrixmatrix_m[$j]} * ${matrixmatrix_percent_nk[$i]} / 100 ]
         k=${n}

         ${TEST_EXE} \
	     --m=${matrixmatrix_m[$j]} \
	     --n=$n \
	     --k=$k \
	     --rep=10 \
	     >> ${TEST_OUTPUT_FILE}
	 progress_bar_step $?
    done
    echo

    #---------------------------------------------------------------------------
    # Create gnuplot data files from results
    #---------------------------------------------------------------------------
    DATA_FILE="${TEST_OUTPUT_FILE%.out}.data"
    grep -E -v '^#|^$' ${TEST_OUTPUT_FILE} > ${DATA_FILE}

    #---------------------------------------------------------------------------
    # Create macros for the latex document
    #---------------------------------------------------------------------------
    MM_SUBFIGURES+=$(cat <<EOF
    \\subfigure[\$n\$ and \$k\$ are ${matrixmatrix_percent_nk[${i}]}\% of \$m\$]{%
      \\label{fig:matrixmatrix-test-results-${i}}
      \\includegraphics[width=0.48\\linewidth]{gb_matrixmatrix_${i}}
    }
EOF
)

      for j in $(seq 0 ${ELEM})
      do
	  k=$[ $j + 1 ]
          MM_TABLE+="$(awk 'NR=='$k' {print $1, "&", $2, "&", $3, "&", $4, "&", $5, "&", $6, "&", $7 }' ${DATA_FILE}) \\\\\\ \n"
      done

      if [ $i -ne $[ ${#matrixmatrix_percent_nk[@]}-1 ] ]; then
          MM_PERCENTSNK+="${matrixmatrix_percent_nk[$i]}, "
          MM_REFSUBFIGURES+=" \\ref{fig:matrixmatrix-test-results-$i}, "
      else
          MM_PERCENTSNK+="and ${matrixmatrix_percent_nk[$i]}\\%"
          MM_REFSUBFIGURES+="and~\\ref{fig:matrixmatrix-test-results-$i}"
      fi

      #-------------------------------------------------------------------------------
      # Create a gnuplot file for this test
      #-------------------------------------------------------------------------------
      GP_FILE="${TEST_OUTPUT_FILE%.out}.gp"
      # Files called from the gnuplot file must be relative to GB_OUPUT_DIR, not to GB_DIR
      REL_DATA_FILE="${DATA_FILE/${GB_OUTPUT_DIR}\/}"
      REL_EPS_FILE="${TEST}/${TEST}_${i}.eps"
      cat > ${GP_FILE}  <<EOF
set xlabel "Number of rows of A (m)"
set ylabel "Time (ms)"
set key bottom right nobox
set grid
set logscale y
set terminal postscript enhanced color
set output "${REL_EPS_FILE}"
plot "${REL_DATA_FILE}" u 1:4 t 'CPU' w linespoints lw 2, "" u 1:5 t 'GPU' w linespoints lw 2
EOF

      # Store the instructions for plotting the GFLOPS of this test 
      PLOTCPU+="\"${REL_DATA_FILE}\" u 1:6 t 'n=k=${matrixmatrix_percent_nk[$i]}% m' w linespoints lw 2"
      PLOTGPU+="\"${REL_DATA_FILE}\" u 1:7 t 'n=k=${matrixmatrix_percent_nk[$i]}% m' w linespoints lw 2"

      if [ $i -ne $[ ${#matrixmatrix_percent_nk[@]} - 1 ] ]
      then
         PLOTCPU+=", "
         PLOTGPU+=", "
      fi

done
#===============================================================================


#===============================================================================
# Create gnuplot file for CPU: Size vs GFLOPS
#-------------------------------------------------------------------------------
GP_FILE="${TEST_OUTPUT_FILE%.out}_GFLOPS_CPU.gp"
# Files called from the gnuplot file must be relative to GB_OUPUT_DIR, not to GB_DIR
REL_EPS_FILE="${TEST}/${TEST}_GFLOPS_CPU.eps"
cat > ${GP_FILE}  <<EOF
set xlabel "Number of rows of A (m)"
set ylabel "CPU GFLOPS"
set key bottom right nobox
set grid
set logscale y
set terminal postscript enhanced color
set output "${REL_EPS_FILE}"
${PLOTCPU}
EOF
#===============================================================================


#===============================================================================
# Create gnuplot file for GPU: Size vs GFLOPS
#-------------------------------------------------------------------------------
GP_FILE="${TEST_OUTPUT_FILE%.out}_GFLOPS_GPU.gp"
# Files called from the gnuplot file must be relative to GB_OUPUT_DIR, not to GB_DIR
REL_EPS_FILE="${TEST}/${TEST}_GFLOPS_GPU.eps"
cat > ${GP_FILE}  <<EOF
set xlabel "Number of rows of A (m)"
set ylabel "GPU GFLOPS"
set key bottom right nobox
set grid
set logscale y
set terminal postscript enhanced color
set output "${REL_EPS_FILE}"
${PLOTGPU}
EOF
#===============================================================================


#===============================================================================
# Write down the macros
#-------------------------------------------------------------------------------
# Sort MM_TABLE first
MM_TABLE=$(echo -e ${MM_TABLE} | sort -n -k 1 -k 2 -k 3 -t "&")
cat >> ${MACROS_FILE} <<EOF
define(MMPERCENTSNK, \`${MM_PERCENTSNK}')
define(MMSUBFIGURES, \`${MM_SUBFIGURES}')
define(MMREFSUBFIGURES, \`${MM_REFSUBFIGURES}')
define(MMTABLE, \`${MM_TABLE}')
EOF
#===============================================================================
