/*==============================================================================
 * gb_matrixvector.cu - measures the computational power of the GPU and the CPU
 *------------------------------------------------------------------------------
 *
 * Copyright 2009-2012 Universitat Jaume I
 *
 * This file is part of GPUBenchmark.
 *
 * GPUBenchmark is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * GPUBenchmark is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GPUBenchmark.  If not, see <http://www.gnu.org/licenses/>.
 *==============================================================================
 */


/*
 * This microbenchmark measures the computational power of the GPU and
 * the CPU using functions from CUBLAS and BLAS libraries,
 * respectively.
 *
 * The measurement is done obtaining the time spent by the operation:
 *
 *   y=alpha*A*x + beta*y,
 *
 * where alpha and beta are two scalars; x and y are two vectors; and
 * A is a matrix.
 *
 * We can do this using gemv routines from CUBLAS (to do the
 * computation in the GPU) and CBLAS (to do the computation in the
 * CPU) libraries.
 *
 * The size of the matrix, and the number of repetitions of the
 * operation are specified by the user at runtime.
 *
 * This test is made both for single-precision and double precision
 * numbers. There are routines in BLAS and CUBLAS supporting both
 * types.
 */


#include "gb_matrixvector.h"



/**
   Runs the matrixvector microbenchmark with the parameters specified
   by the supplied options.

   @param [in]  argc       Count of the arguments supplied to the program.
   @param [in]  **argv     Array of pointers to the strings which are those arguments.
   @return 0 if the test passed correctly, other number otherwise.
*/
int main(int argc, char** argv)
{
  struct options_t options; // Command line options struct
  int i;
  int sizeofelem;
  double *time_gpu;
  double *time_cpu;
  double mean_time_gpu;
  double mean_time_cpu;

  // Get options from command line
  getopts(argc, argv, &options);

  if (options.help)
    matrixvectorHelp();

  if (options.device_info)
    showProperties(options.output, options.device);

  // Check required parameters
  if (options.rows==-1 || options.cols==-1)
    error("The number of rows and columns must be specified.");

  // Set sizeofelem
  if (options.use_simple)
    sizeofelem=sizeof(float);
  else
    sizeofelem=sizeof(double);

  // Time vectors
  time_gpu=(double *) malloc(options.rep*sizeof(double));
  time_cpu=(double *) malloc(options.rep*sizeof(double));

  // Display matrix and vectors sizes
  if (options.verbosity)
    fprintf(options.output,
	    "\n"
	    "Size of matrix \"A\": %8.2f KiB\n"
	    "Size of vector \"x\": %8.2f KiB\n"
	    "Size of vector \"y\": %8.2f KiB\n\n",
	    (options.rows*options.cols*sizeofelem)/1024.0,
	    (options.cols*sizeofelem)/1024.0,
	    (options.rows*sizeofelem)/1024.0);

  // Run the tests options.rep times
  for (i=0; i<options.rep; i++)
    {
      if (matrixvector(options.device, options.rows, options.cols, sizeofelem, &time_cpu[i], &time_gpu[i]) == -1)
	error("The test %d/%d failed.", i+1, options.rep);
    }

  // Compute mean cpu and gpu times
  mean_time_cpu=mean(time_cpu, options.rep);
  mean_time_gpu=mean(time_gpu, options.rep);

  // @warning: '2.0' is used instead of '2' to overcome integer
  //           overflow when computing Gflops
  if (options.verbosity)
    {
      fprintf(options.output,
	      "CPU time (CBLAS):  %7.4f ms\n"
	      "GPU time (CUBLAS): %7.4f ms\n"
	      "GFLOPS CPU: %11.4f\n"
	      "GFLOPS GPU: %11.4f\n",
	      mean_time_cpu*1000.0,
	      mean_time_gpu*1000.0,
	      2.0*options.rows*options.cols*1.0e-9/mean_time_cpu,
	      2.0*options.rows*options.cols*1.0e-9/mean_time_gpu);
    }
  else
    {
      fprintf(options.output,
	      //1234567.1234567.1234567.1234567890123.1234567890123.1234567890123.1234567890123.
	      "#   M       N      CBLAS (ms)   CUBLAS (ms)   GFLOPS CPU    GFLOPS GPU\n"
	      "#-------+-------+-------------+-------------+-------------+-------------+\n"
	      " %7d %7d %13.2f %13.2f %13.2f %13.2f\n"
	      "",
	      options.rows, options.cols,
	      mean_time_cpu*1000.0,
	      mean_time_gpu*1000.0,
	      2.0*options.rows*options.cols*1.0e-9/mean_time_cpu,
	      2.0*options.rows*options.cols*1.0e-9/mean_time_gpu);
    }

  // Exit main
  fprintf(options.output, "\n");
  fclose(options.output);
  free(time_cpu);
  free(time_gpu);
  cudaDeviceReset();
  return 0;

}


/**
   Measures the time spent by the CPU and the GPU to do the operation
   \f$y=alpha*A*x+beta*y\f$, using the Sgemv() or Dgemv() functions of
   CBLAS and CUBLAS libraries, respectively.

   \em A is a matrix, \em x and \em y are vectors, and \em alpha and
   \em beta are scalars.

   @param [in]  device       GPU device that is going to be used.
   @param [in]  rows         The number of rows of matrix A (and elements of y).
   @param [in]  cols         The number of columns of matrix A (and elements x).
   @param [in]  sizeofelem   The size in bytes of each matrix element.
   @param [out] time_cpu     Time spent by CPU to do the operation (in seconds).
   @param [out] time_gpu     Time spent by GPU to do the operation (in seconds).
   @return 0 if everything went ok, other number otherwise.
*/
int matrixvector(int device, int rows, int cols, int sizeofelem, double *time_cpu, double *time_gpu)
{
  cudaError_t cudaStat;
  cublasStatus_t status;
  cublasHandle_t handle;
  bool error=false;
  int i, j;

  float s_alpha=2.0;
  float s_beta=3.0;
  double d_alpha=2.0;
  double d_beta=3.0;
  void *A_mm,  *x_mm,  *y_mm;
  void *A_gpu, *x_gpu, *y_gpu;
  void *y_result;

  double norm_y_result, norm_y_mm, result;

  cudaEvent_t start, end;
  float float_time;

  struct timeval t0, t1;
  struct timezone z;


  // Allocate space in main memory for CPU matrices
  A_mm= (void *) malloc(rows*cols*sizeofelem);
  x_mm= (void *) malloc(cols*sizeofelem);
  y_mm= (void *) malloc(rows*sizeofelem);
  y_result= (void *) malloc(rows*sizeofelem);
  if (!A_mm || !x_mm || !y_mm || !y_result)
    {
      no_exit_error("Allocation of matrix and vectors in main memory failed.");
      free(A_mm);
      free(x_mm);
      free(y_mm);
      free(y_result);
      return -1;
    }

  // Fill matrix A, and vectors x and y with arbitrary values (column-major indexing)
  if (sizeofelem==sizeof(float))
    {
      for (j=0;j<cols; j++)
	for (i=0;i<rows;i++)
	  ((float*)A_mm)[IDX2C(i,j,rows)]= j*rows+i;
      for (j=0; j<cols; j++)
	((float*)x_mm)[j]= j;
      for (i=0;i<rows;i++)
	((float*)y_mm)[i]= i;
    }
  else
    {
      for (j=0;j<cols; j++)
	for (i=0; i<rows; i++)
	  ((double*)A_mm)[IDX2C(i,j,rows)]= j*rows+i;
      for (j=0; j<cols; j++)
	((double*)x_mm)[j]= j;
      for (i=0; i<rows; i++)
	((double*)y_mm)[i]= i;
    }

  // Set CUDA device
  cudaSetDevice(device);

  // Allocate space in GPU memory for the matrix and the vectors
  cudaStat = cudaMalloc((void**)&A_gpu, rows*cols*sizeofelem);
  if (cudaStat != cudaSuccess)
    {
      no_exit_error("Device memory allocation failed for matrix \"A\".");
      free(A_mm);
      free(x_mm);
      free(y_mm);
      free(y_result);
      return -1;
    }
  cudaStat = cudaMalloc((void**)&x_gpu, cols*sizeofelem);
  if (cudaStat != cudaSuccess)
    {
      no_exit_error("Device memory allocation failed for vector \"x\".");
      free(A_mm);
      free(x_mm);
      free(y_mm);
      free(y_result);
      cudaFree(A_gpu);
      return -1;
    }
  cudaStat = cudaMalloc((void**)&y_gpu, rows*sizeofelem);
  if (cudaStat != cudaSuccess)
    {
      no_exit_error("Device memory allocation failed for vector \"y\".");
      free(A_mm);
      free(x_mm);
      free(y_mm);
      free(y_result);
      cudaFree(A_gpu);
      cudaFree(x_gpu);
      return -1;
    }

  // Initialization of CUBLAS
  status = cublasCreate(&handle);
  if (status != CUBLAS_STATUS_SUCCESS)
    {
      no_exit_error("CUBLAS initialization failed.");
      return -1;
    }

  // Copy the matrix and the vectors from main memory to GPU
  status= cublasSetMatrix(rows, cols, sizeofelem, A_mm, rows, A_gpu, rows);
  if(status != CUBLAS_STATUS_SUCCESS)
    {
      no_exit_error("Copying of matrix \"A\" to GPU memory failed.");
      error=true;
    }
  status = cublasSetVector(cols, sizeofelem, x_mm, 1, x_gpu, 1);
  if (status != CUBLAS_STATUS_SUCCESS)
    {
      no_exit_error("Copying of vector \"x\" to GPU memory failed.");
      error=true;
    }
  status = cublasSetVector(rows, sizeofelem, y_mm, 1, y_gpu, 1);
  if (status != CUBLAS_STATUS_SUCCESS)
    {
      no_exit_error("Copying of vector \"y\" to GPU memory failed.");
      error=true;
    }
  if (error)
    {
      free(A_mm);
      free(x_mm);
      free(y_mm);
      free(y_result);
      cudaFree(A_gpu);
      cudaFree(x_gpu);
      cudaFree(y_gpu);
      return -1;
    }

  // Create start and end events
  cudaEventCreate(&start);
  cudaEventCreate(&end);


  // ======================================================================
  // CUBLAS operation
  // ----------------------------------------------------------------------

  // Record "start" event
  cudaDeviceSynchronize();
  cudaEventRecord(start, 0);

  // Do the gemv operation on GPU using cublas<T>gemv function
  if (sizeofelem==sizeof(float))
    status=cublasSgemv(handle, CUBLAS_OP_N, rows, cols,
		       &s_alpha,  (float*)A_gpu, rows, (float*)x_gpu, 1, &s_beta, (float*)y_gpu, 1);
  else
    status=cublasDgemv(handle, CUBLAS_OP_N, rows, cols,
		       &d_alpha,  (double*)A_gpu, rows, (double*)x_gpu, 1, &d_beta, (double*)y_gpu, 1);
  if (status != CUBLAS_STATUS_SUCCESS)
    {
      if (sizeofelem==sizeof(float))
	no_exit_error("The cublasSgemv function failed.");
      else
	no_exit_error("The cublasDgemv function failed.");
      free(A_mm);
      free(x_mm);
      free(y_mm);
      free(y_result);
      cudaFree(A_gpu);
      cudaFree(x_gpu);
      cudaFree(y_gpu);
      cublasDestroy(handle);
      cudaEventDestroy(start);
      cudaEventDestroy(end);
      return -1;
    }

  // Record "end" event and synchronize
  cudaEventRecord(end, 0);
  cudaEventSynchronize(end);

  // Get the elapsed time (in ms) between "start" and "end" events
  cudaEventElapsedTime(&float_time, start, end);

  // Convert time to seconds and store it in time_gpu
  *time_gpu=(double)float_time/1000.0;

  // ======================================================================


  // Copy the result from GPU memory space to main memory space
  status = cublasGetVector(rows, sizeofelem, y_gpu, 1, y_result, 1);
  if (status != CUBLAS_STATUS_SUCCESS)
    {
      no_exit_error("Copying of vector \"y\" to CPU memory failed.");
      free(A_mm);
      free(x_mm);
      free(y_mm);
      free(y_result);
      cudaFree(A_gpu);
      cudaFree(x_gpu);
      cudaFree(y_gpu);
      cublasDestroy(handle);
      cudaEventDestroy(start);
      cudaEventDestroy(end);
      return -1;
    }


  // ======================================================================
  // CBLAS operation
  // ----------------------------------------------------------------------

  // Capture current time of day (for CBLAS <T>gemv)
  gettimeofday(&t0, &z);

  // Do the gemm operation on CPU using cblas_<T>gemv function
  if (sizeofelem==sizeof(float))
    cblas_sgemv(CblasColMajor, CblasNoTrans, rows, cols,
		s_alpha, (float*)A_mm, rows, (float*)x_mm, 1, s_beta, (float*)y_mm, 1);
  else
    cblas_dgemv(CblasColMajor, CblasNoTrans, rows, cols,
		d_alpha, (double*)A_mm, rows, (double*)x_mm, 1, d_beta, (double*)y_mm, 1);

  // Capture time of day after doing the CBLAS <T>gemm operation
  gettimeofday(&t1, &z);

  // Compute the time difference between t1 and t0
  *time_cpu = timeval_diff(t1, t0);

  // ======================================================================


  // Compare the results obtained using CBLAS (CPU) and CUBLAS (GPU)
  if (sizeofelem==sizeof(float))
    {
      for(i=0;i<rows;i++)
	((float*)y_mm)[i]=((float*)y_result)[i]-((float*)y_mm)[i];
      norm_y_result=cblas_snrm2(rows, (float*)y_result, 1);
      norm_y_mm=cblas_snrm2(rows, (float*)y_mm, 1);
    }
  else
    {
      for(i=0;i<rows;i++)
	((double*)y_mm)[i]=((double*)y_result)[i]-((double*)y_mm)[i];
      norm_y_result=cblas_dnrm2(rows, (double*)y_result, 1);
      norm_y_mm=cblas_dnrm2(rows, (double*)y_mm, 1);
    }
  if (norm_y_result == 0.0)
    result= norm_y_mm;
  else
    result= norm_y_mm/norm_y_result;
  if ( (sizeofelem==sizeof(float) and result > FLT_EPSILON*10) or (sizeofelem==sizeof(double) and result > DBL_EPSILON*10 ) )
    warning("The difference between the results obtained with CBLAS and CUBLAS are too significative.");


  // Free the memories
  free(A_mm);
  free(x_mm);
  free(y_mm);
  free(y_result);
  cudaFree(A_gpu);
  cudaFree(x_gpu);
  cudaFree(y_gpu);

  // Destroy CUBLAS handle
  cublasDestroy(handle);

  // Destroy "start" and "end" events
  cudaEventDestroy(start);
  cudaEventDestroy(end);

  // Return
  return 0;
}


/**
   Show usage information
*/
void matrixvectorHelp(void)
{
  printf("Usage: ./gb_matrixvector [OPTION]..\n"
	 "Measures the performance of gpu and cpu for the matrix vector\n"
	 "multiplication operation using CUBLAS and CBLAS, respectivelly.\n"
	 "\n"
	 "Required options:\n"
	 "      --rows=NUM       number of rows of the matrix\n"
	 "      --cols=NUM       number of columns of the matrix\n"
	 "\n"
	 "Other options:\n"
	 "      --device=NUM     use  the CUDA device with id=NUM (default: 0)\n"
	 "      --device-info    show information about the given device and exit\n"
	 "      --double         run the test using double precision elements\n"
	 "      --rep=NUM        number of repetitions of the test (default: 10)\n"
	 "  -v, --verbose        be more verbose\n"
	 "      --dnfo           show information about the used device\n"
	 "  -o, --output=FNAME   send the output to the specified file\n"
	 "      --help           display this help and exit\n"
	 "\n");
  exit(0);
}
