#!/bin/bash

#===============================================================================
# gb_matrixvector_run.sh
#-------------------------------------------------------------------------------
#
# Copyright 2009-2012 Universitat Jaume I
#
# This file is part of GPUBenchmark.
#
# GPUBenchmark is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GPUBenchmark is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GPUBenchmark.  If not, see <http://www.gnu.org/licenses/>.
#===============================================================================


#===============================================================================
# Tests parameters
#-------------------------------------------------------------------------------
# matrixvector
#  - percent_cols are the % of rows
matrixvector_rows=(128 256 512 1024 2048 4096 8192 10240)
matrixvector_percent_cols=(25 50 75 100)
#===============================================================================


#===============================================================================
# Check that we are run from gb_run.sh
#-------------------------------------------------------------------------------
if [ -z "${GB_DIR}" ]; then
    echo "Error: this file must be called from 'gb_run.sh'"
fi
#===============================================================================


#===============================================================================
# matrixvector tests
#-------------------------------------------------------------------------------
echo "Running matrix per vector tests..."

#-------------------------------------------------------------------------------
# Checks and initialization of variables
#-------------------------------------------------------------------------------
ELEM=$[ ${#matrixvector_rows[@]} - ${ONE} ]
TEST=gb_matrixvector
TEST_EXE="${TEST}/${TEST}"
TEST_OUTPUT_DIR="${GB_OUTPUT_DIR}/${TEST}"
[ -d "${TEST_OUTPUT_DIR}" ] || mkdir -p "${TEST_OUTPUT_DIR}"

PLOTCPU="plot "
PLOTGPU="plot "
MV_PERCENTS=""
MV_SUBFIGURES=""
MV_REFSUBFIGURES=""
MV_TABLE=""

for i in $(seq 0 $[ ${#matrixvector_percent_cols[@]} - 1] )
do
    TEST_OUTPUT_FILE="${TEST_OUTPUT_DIR}/${TEST}_${i}.out"
    [ -e "${TEST_OUTPUT_FILE}" ] && rm -f "${TEST_OUTPUT_FILE}"

    #---------------------------------------------------------------------------
    # Actual running of test
    #---------------------------------------------------------------------------
    echo -n "Running matrix per vector test ($[ ${i} + 1 ]/${#matrixvector_percent_cols[@]}): "
    progress_bar ${ELEM}
    for j in $(seq 0 ${ELEM})
    do
         cols=$[ ${matrixvector_rows[$j]} * ${matrixvector_percent_cols[$i]} / 100 ]

         ${TEST_EXE} \
	     --rows=${matrixvector_rows[$j]} \
	     --cols=${cols} \
	     --rep=10 \
	     >> ${TEST_OUTPUT_FILE}
	 progress_bar_step $?
    done
    echo

    #---------------------------------------------------------------------------
    # Create gnuplot data files from results
    #---------------------------------------------------------------------------
    DATA_FILE="${TEST_OUTPUT_FILE%.out}.data"
    grep -E -v '^#|^$' ${TEST_OUTPUT_FILE} > ${DATA_FILE}

    #---------------------------------------------------------------------------
    # Create macros for the latex document
    #---------------------------------------------------------------------------
    MV_SUBFIGURES+=$(cat <<EOF
    \\subfigure[\$n\$ is ${matrixvector_percent_cols[${i}]}\% of \$m\$]{ %
      \\label{fig:matrixvector-test-results-${i}}
      \\includegraphics[width=0.48\\linewidth]{gb_matrixvector_${i}}
    }
EOF
)

    for j in $(seq 0 ${ELEM})
    do
	k=$[ $j + 1 ]
        MV_TABLE+="$(awk 'NR=='$k' {print $1, "&", $2, "&", $3, "&", $4, "&", $5, "&", $6 }' ${DATA_FILE}) \\\\\\ \n"
    done

    if [ $i -ne $[ ${#matrixvector_percent_cols[@]}-1 ] ]; then
        MV_PERCENTS+="${matrixvector_percent_cols[$i]}, "
        MV_REFSUBFIGURES+=" \\ref{fig:matrixvector-test-results-$i}, "
    else
        MV_PERCENTS+="and ${matrixvector_percent_cols[$i]}\\%"
        MV_REFSUBFIGURES+="and~\\ref{fig:matrixvector-test-results-$i}"
    fi

      #-------------------------------------------------------------------------------
      # Create a gnuplot file for this test
      #-------------------------------------------------------------------------------
      GP_FILE="${TEST_OUTPUT_FILE%.out}.gp"
      # Files called from the gnuplot file must be relative to GB_OUPUT_DIR, not to GB_DIR
      REL_DATA_FILE="${DATA_FILE/${GB_OUTPUT_DIR}\/}"
      REL_EPS_FILE="${TEST}/${TEST}_${i}.eps"
      cat > ${GP_FILE}  <<EOF
set xlabel "Number of rows of A (m)"
set ylabel "Time (ms)"
set key bottom right nobox
set grid
set logscale y
set terminal postscript enhanced color
set output "${REL_EPS_FILE}"
plot "${REL_DATA_FILE}" u 1:3 t 'CPU' w linespoints lw 2, "" u 1:4 t 'GPU' w linespoints lw 2
EOF

      # Store the instructions for plotting the GFLOPS of this test 
      PLOTCPU+="\"${REL_DATA_FILE}\" u 1:5 t 'n=${matrixvector_percent_cols[$i]}% m' w linespoints lw 2"
      PLOTGPU+="\"${REL_DATA_FILE}\" u 1:6 t 'n=${matrixvector_percent_cols[$i]}% m' w linespoints lw 2"

      if [ $i -ne $[ ${#matrixvector_percent_cols[@]} - 1 ] ]
      then
         PLOTCPU+=", "
         PLOTGPU+=", "
      fi

done
#===============================================================================


#===============================================================================
# Create gnuplot file for CPU: Size vs GFLOPS
#-------------------------------------------------------------------------------
GP_FILE="${TEST_OUTPUT_FILE%.out}_GFLOPS_CPU.gp"
# Files called from the gnuplot file must be relative to GB_OUPUT_DIR, not to GB_DIR
REL_EPS_FILE="${TEST}/${TEST}_GFLOPS_CPU.eps"
cat > ${GP_FILE}  <<EOF
set xlabel "Number of rows of A (m)"
set ylabel "CPU GFLOPS"
set key bottom right nobox
set grid
set logscale y
set terminal postscript enhanced color
set output "${REL_EPS_FILE}"
${PLOTCPU}
EOF
#===============================================================================


#===============================================================================
# Create gnuplot file for GPU: Size vs GFLOPS
#-------------------------------------------------------------------------------
GP_FILE="${TEST_OUTPUT_FILE%.out}_GFLOPS_GPU.gp"
# Files called from the gnuplot file must be relative to GB_OUPUT_DIR, not to GB_DIR
REL_EPS_FILE="${TEST}/${TEST}_GFLOPS_GPU.eps"
cat > ${GP_FILE}  <<EOF
set xlabel "Number of rows of A (m)"
set ylabel "GPU GFLOPS"
set key bottom right nobox
set grid
set logscale y
set terminal postscript enhanced color
set output "${REL_EPS_FILE}"
${PLOTGPU}
EOF
#===============================================================================


#===============================================================================
# Write down the macros
#-------------------------------------------------------------------------------
# Sort MV_TABLE first
MV_TABLE=$(echo -e ${MV_TABLE} | sort -n -k 1 -k 2 -t "&")
cat >> ${MACROS_FILE} <<EOF
define(MVPERCENTS, \`${MV_PERCENTS}')
define(MVSUBFIGURES, \`${MV_SUBFIGURES}')
define(MVREFSUBFIGURES, \`${MV_REFSUBFIGURES}')
define(MVTABLE, \`${MV_TABLE}')
EOF
#===============================================================================
