#!/bin/bash

#===============================================================================
# gb_run.sh
#-------------------------------------------------------------------------------
#
# Copyright 2009-2012 Universitat Jaume I
#
# This file is part of GPUBenchmark.
#
# GPUBenchmark is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GPUBenchmark is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GPUBenchmark.  If not, see <http://www.gnu.org/licenses/>.
#===============================================================================


#===============================================================================
# Tests parameters
#-------------------------------------------------------------------------------
#@todo: DEVICE parameter to select the GPU device to be used by the tests
#@todo: REPETITIONS to change the number of repetitions of each test
#===============================================================================


#===============================================================================
# Initialization and checks
#-------------------------------------------------------------------------------
export GB_DIR=$(dirname $0)
export GB_OUTPUT_DIR="gpubenchmark_${HOSTNAME}_results"
export MACROS_FILE="${GB_OUTPUT_DIR}/macros.txt"

#POTENTIAL_TESTS=$(find ${GB_DIR} -maxdepth 1 -name "gb_*" -type d -exec basename {} \;)
POTENTIAL_TESTS=$(cat ${GB_DIR}/CMakeLists.txt   \
    | grep -v "#" | grep "add_subdirectory(gb_"  \
    | sed -e 's/add_subdirectory(\([^)]*\))/\1/'   )
TESTS=""
for TEST in ${POTENTIAL_TESTS}; do
    EXE="${GB_DIR}/${TEST}/${TEST}"
    if [ -x "${EXE}" ]; then
	TESTS="${TESTS}${TEST} "
    fi
done

if [ -z "${TESTS}" ]; then
    cat >&2 <<EOF
$0: No tests found.

Please, compile GPUBenchmark before running this script.
EOF
    exit -1
fi


# Go into GB_DIR and check GB_OUTPUT_DIR
cd ${GB_DIR}
[ -d "${GB_OUTPUT_DIR}" ] || mkdir -p "${GB_OUTPUT_DIR}"

# Put common macros in MACROS_FILE
echo "define(HOSTNAME,${HOSTNAME})" > "${MACROS_FILE}"
echo "define(TEST_INPUTS," >> "${MACROS_FILE}"
for TEST in ${TESTS}; do
    echo "\\input{${TEST}.tex}" >> "${MACROS_FILE}"
done
echo ")" >> "${MACROS_FILE}"
#===============================================================================


#===============================================================================
# progress_bar and progress_bar_step functions
#-------------------------------------------------------------------------------
progress_bar() {
    echo -n "["
    for in in $(seq 0 $1); do
	echo -n "-"
    done
    echo -n "]"
    for in in $(seq 0 $[ $1 + 1 ]); do
	echo -ne "\b"
    done    
}
progress_bar_step() {
    if [ -z "$1" -o "$1" -eq 0 ]; then
	echo -n "*"
    else
	echo -n "!"
    fi
}
export -f progress_bar
export -f progress_bar_step
#===============================================================================


#===============================================================================
# CPU and GPU info
#-------------------------------------------------------------------------------
echo "Writing down cpu and gpu info..."
cat /proc/cpuinfo | \
    grep -E 'processor|model name|cpu MHz|cache size|cpu cores|bogomips' \
    | sed -e 's/  */ /g' -e 's/\t//g' -e 's/ : /:/g' -e 's/: /:/g' \
    | awk -F ":" \
       'NR==1 {next} /processor/ {print ""; next} /model name/ {print $2; next} {printf(" %-10s : %s\n", $1, $2);}' \
    > ${GB_OUTPUT_DIR}/cpu_info.txt
./gb_gputoram/gb_gputoram --device=0 --device-info \
    > ${GB_OUTPUT_DIR}/gpu_info.txt
echo ${HOSTNAME} > ${GB_OUTPUT_DIR}/hostname.txt
#===============================================================================


#===============================================================================
# Copy all the LaTeX templates and related files to the output directory
#-------------------------------------------------------------------------------
echo "Copying LaTeX templates to ${GB_OUTPUT_DIR}/templates/..."
[ -d ${GB_OUTPUT_DIR}/templates/ ] || mkdir ${GB_OUTPUT_DIR}/templates/
rm -f ${GB_OUTPUT_DIR}/templates/*_tpl.tex
TPL_FILES=$(find ./ -wholename "./gb_*/*_tpl.tex" -o \
                    -wholename "./templates/*_tpl.tex" )
cp ${TPL_FILES} ${GB_OUTPUT_DIR}/templates/
echo "Copying 'gb_make_pdf' to ${GB_OUTPUT_DIR}/..."
cp ./templates/gb_make_pdf.sh ${GB_OUTPUT_DIR}/
#===============================================================================


#===============================================================================
# Run all the tests
#-------------------------------------------------------------------------------
export ONE=1 # Use 5 instead of 1 when checking the software
for TEST in ${TESTS}; do
     "./${TEST}/${TEST}_run.sh"
done
#===============================================================================


# The end!
cat <<EOF

The GPUBenchmark tests have ended. You can now either run the script:
'${GB_DIR}/${GB_OUTPUT_DIR}/gb_make_pdf.sh'; or copy the contents of
the '${GB_DIR}/${GB_OUTPUT_DIR}/' directory to another computer and
run the 'gb_make_pdf.sh' script there.

EOF

exit 0
